import pytest
from unittest.mock import MagicMock, patch
from pigeon.client import Pigeon
from pydantic import ValidationError


@pytest.fixture
def pigeon_client():
    with patch("pigeon.utils.setup_logging") as mock_logging:
        with patch("pigeon.client.stomp"):
            client = Pigeon(
                "test",
                host="localhost",
                port=61613,
                logger=mock_logging.Logger(),
                load_topics=False,
            )
            yield client


def test_one_arg(pigeon_client):
    mock_stomp_message = MagicMock()
    mock_stomp_message.headers = {
        "subscription": "test.msg",
        "hash": "abcd",
    }

    mock_message = MagicMock()

    def callback(msg):
        mock_message.deserialize.assert_called_with(mock_stomp_message.body)
        assert msg == mock_message.deserialize()

    pigeon_client._connection = MagicMock()
    pigeon_client._topics["test.msg"] = mock_message
    pigeon_client._hashes["test.msg"] = "abcd"
    pigeon_client.subscribe("test.msg", callback)

    pigeon_client._handle_message(mock_stomp_message)


def test_two_args(pigeon_client):
    mock_stomp_message = MagicMock()
    mock_stomp_message.headers = {
        "subscription": "test.msg",
        "hash": "abcde",
    }

    mock_message = MagicMock()

    def callback(msg, topic):
        mock_message.deserialize.assert_called_with(mock_stomp_message.body)
        assert msg == mock_message.deserialize()
        assert topic == "test.msg"

    pigeon_client._connection = MagicMock()
    pigeon_client._topics["test.msg"] = mock_message
    pigeon_client._hashes["test.msg"] = "abcde"
    pigeon_client.subscribe("test.msg", callback)

    pigeon_client._handle_message(mock_stomp_message)


def test_three_args(pigeon_client):
    mock_stomp_message = MagicMock()
    mock_stomp_message.headers = {
        "subscription": "test.msg",
        "hash": "123abc",
    }

    mock_message = MagicMock()

    def callback(msg, topic, headers):
        mock_message.deserialize.assert_called_with(mock_stomp_message.body)
        assert msg == mock_message.deserialize()
        assert topic == "test.msg"
        assert headers == mock_stomp_message.headers

    pigeon_client._connection = MagicMock()
    pigeon_client._topics["test.msg"] = mock_message
    pigeon_client._hashes["test.msg"] = "123abc"
    pigeon_client.subscribe("test.msg", callback)

    pigeon_client._handle_message(mock_stomp_message)


def test_var_args(pigeon_client):
    mock_stomp_message = MagicMock()
    mock_stomp_message.headers = {
        "subscription": "test.msg",
        "hash": "xyz987",
    }

    mock_message = MagicMock()

    def callback(*args):
        mock_message.deserialize.assert_called_with(mock_stomp_message.body)
        assert len(args) == 3
        assert args[0] == mock_message.deserialize()
        assert args[1] == "test.msg"
        assert args[2] == mock_stomp_message.headers

    pigeon_client._connection = MagicMock()
    pigeon_client._topics["test.msg"] = mock_message
    pigeon_client._hashes["test.msg"] = "xyz987"
    pigeon_client.subscribe("test.msg", callback)

    pigeon_client._handle_message(mock_stomp_message)


def create_mock_message(body="", **headers):
    return MagicMock(body=body, headers=headers)


def test_topic_does_not_exist(pigeon_client):
    mock_message = create_mock_message(subscription="not.a.real.message")

    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        "Received a message on an unregistered topic: not.a.real.message"
    )


def test_hash_mismatch(pigeon_client):
    mock_message = create_mock_message(subscription="test", hash="abc1")

    pigeon_client._topics["test"] = None
    pigeon_client._hashes["test"] = "abcd"
    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        "Received a message on topic 'test' with an incorrect hash: abc1. Expected: abcd"
    )


def test_validation_error(pigeon_client):
    mock_message = create_mock_message(subscription="test", hash="abc123")
    mock_msg_def = MagicMock()
    mock_msg_def.deserialize.side_effect = ValidationError.from_exception_data(
        title="Test", line_errors=[]
    )

    pigeon_client._topics["test"] = mock_msg_def
    pigeon_client._hashes["test"] = "abc123"
    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        "Failed to deserialize message on topic 'test' with error:\n0 validation errors for Test\n"
    )


def test_no_callback(pigeon_client):
    mock_message = create_mock_message(subscription="test", hash="4321")

    pigeon_client._topics["test"] = MagicMock()
    pigeon_client._hashes["test"] = "4321"
    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        "No callback for message received on topic 'test'."
    )


def test_bad_signature(pigeon_client):
    mock_message = create_mock_message(subscription="test", hash="lmnop")
    callback = lambda a, b, c, d: None

    pigeon_client._topics["test"] = MagicMock()
    pigeon_client._hashes["test"] = "lmnop"
    pigeon_client.subscribe("test", callback)
    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        f"Callback signature for topic 'test' not acceptable. Call failed with error:\nFunction '{callback}' requires 4 positional arguments, but only 3 are available."
    )


def test_callback_exception(pigeon_client):
    mock_message = create_mock_message(subscription="test", hash="987654321")

    pigeon_client._topics["test"] = MagicMock()
    pigeon_client._hashes["test"] = "987654321"
    pigeon_client.subscribe(
        "test", MagicMock(side_effect=RecursionError("This is a test error."))
    )
    pigeon_client._handle_message(mock_message)

    pigeon_client._logger.warning.assert_called_with(
        f"Callback for topic 'test' failed with error:", exc_info=True
    )



from pigeon import Pigeon, BaseMessage
import pytest
from stomp.utils import Frame
from py_zipkin.request_helpers import extract_zipkin_attrs_from_headers
from py_zipkin.util import ZipkinAttrs


class Msg(BaseMessage):
    value: str


@pytest.mark.parametrize(
    "use_zipkin, zipkin_headers, message_headers",
    [
        (
            False,
            {
                "X-B3-TraceId": "trace_id",
                "X-B3-SpanId": "span_id",
                "X-B3-ParentSpanId": "parent_id",
                "X-B3-Sampled": "sample!",
                "X-B3-Flags": "wave!",
            },
            {
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
            },
        ),
        (
            True,
            {
                "X-B3-TraceId": "trace_id",
                "X-B3-SpanId": "span_id",
                "X-B3-ParentSpanId": "parent_id",
                "X-B3-Sampled": "sample!",
                "X-B3-Flags": "wave!",
            },
            {
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
                "X-B3-TraceId": "trace_id",
                "X-B3-SpanId": "span_id",
                "X-B3-ParentSpanId": "parent_id",
                "X-B3-Sampled": "sample!",
                "X-B3-Flags": "wave!",
            },
        ),
        (
            True,
            {},
            {
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
            },
        ),
    ],
)
def test_send_zipkin(mocker, use_zipkin, zipkin_headers, message_headers):
    mocker.patch(
        "pigeon.client.create_http_headers_for_new_span", return_value=zipkin_headers
    )
    mocker.patch("pigeon.client.setup_zipkin_transport", return_value=True)
    mocker.patch("pigeon.client.get_str_time_ms", return_value="0")
    mocker.patch("pigeon.client.stomp.Connection12")

    client = Pigeon("the service", send_zipkin_headers=use_zipkin, load_topics=False)
    client._name = "a name"
    client._hostname = "hosty mchostface"
    client._pid = 1

    client.register_topic("test", Msg)
    client._hashes["test"] = "abc123"

    client.send("test", value="something")

    client._connection.send.assert_called_with(
        destination="test", body='{"value":"something"}', headers=message_headers
    )


@pytest.mark.parametrize(
    "use_zipkin, message_headers, span_created",
    [
        (
            False,
            {
                "X-B3-TraceId": "trace_id",
                "X-B3-SpanId": "span_id",
                "X-B3-ParentSpanId": "parent_id",
                "X-B3-Sampled": "true",
                "X-B3-Flags": "0",
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
                "subscription": "test",
            },
            False,
        ),
        (
            False,
            {
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
                "subscription": "test",
            },
            False,
        ),
        (
            True,
            {
                "X-B3-TraceId": "trace_id",
                "X-B3-SpanId": "span_id",
                "X-B3-ParentSpanId": "parent_id",
                "X-B3-Sampled": "true",
                "X-B3-Flags": "0",
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
                "subscription": "test",
            },
            True,
        ),
        (
            True,
            {
                "source": "a name",
                "service": "the service",
                "hostname": "hosty mchostface",
                "pid": 1,
                "hash": "abc123",
                "sent_at": "0",
                "subscription": "test",
            },
            False,
        ),
    ],
)
def test_receive_zipkin(mocker, use_zipkin, message_headers, span_created):
    mocker.patch("pigeon.client.setup_zipkin_transport")
    mocker.patch("pigeon.client.stomp.Connection12")
    zipkin_span = mocker.patch("pigeon.client.zipkin_span")

    client = Pigeon("the service", create_zipkin_spans=use_zipkin, load_topics=False)

    client.register_topic("test", Msg)
    client._hashes["test"] = "abc123"

    callback = mocker.MagicMock()
    client.subscribe("test", callback)

    msg = Frame("", body='{"value":"something"}', headers=message_headers)

    client._handle_message(msg)

    callback.assert_called_with(Msg(value="something"), "test", message_headers)

    if span_created:
        zipkin_span.assert_called_with(
            service_name=client._service,
            span_name="handle test",
            transport_handler=client._zipkin_transport,
            zipkin_attrs=ZipkinAttrs("trace_id", "span_id", "parent_id", "0", True),
        )
        zipkin_span().__enter__.assert_called()
        zipkin_span().__exit__.assert_called()
    else:
        zipkin_span.assert_not_called()



import pytest
from unittest.mock import MagicMock, patch
from stomp.exception import ConnectFailedException
from pigeon.client import Pigeon
from pigeon.exceptions import NoSuchTopicException
from pigeon import BaseMessage


class MockMessage(BaseMessage):
    field1: str


@pytest.fixture
def pigeon_client():
    with patch("pigeon.utils.setup_logging") as mock_logging:
        topics = {"topic1": MockMessage, "topic2": MockMessage}
        client = Pigeon(
            "test",
            host="localhost",
            port=61613,
            logger=mock_logging.Logger(),
            load_topics=False,
        )
        client.register_topics(topics)
        yield client


@pytest.mark.parametrize(
    "username, password, expected_log",
    [
        (None, None, "Connected to STOMP server."),
        ("user", "pass", "Connected to STOMP server."),
    ],
    ids=["no-auth", "with-auth"],
)
def test_connect(pigeon_client, username, password, expected_log):
    # Arrange
    pigeon_client._connection.connect = MagicMock()
    pigeon_client.subscribe = MagicMock()
    pigeon_client._announce = MagicMock()

    # Act
    pigeon_client.connect(username=username, password=password)

    # Assert
    pigeon_client._connection.connect.assert_called_with(
        username=username, passcode=password, wait=True
    )
    pigeon_client._logger.info.assert_called_with(expected_log)
    pigeon_client.subscribe.assert_called_with(
        "&_request_state", pigeon_client._update_state
    )
    pigeon_client._announce.assert_called()


@pytest.mark.parametrize(
    "username, password",
    [
        (None, None),
        ("user", "pass"),
    ],
    ids=["connect-fail-no-auth", "connect-fail-with-auth"],
)
def test_connect_failure(pigeon_client, username, password):
    # Arrange
    pigeon_client._connection.connect = MagicMock(
        side_effect=ConnectFailedException("Connection failed")
    )
    retry_limit = 1
    pigeon_client._logger.error = MagicMock()

    # Act & Assert
    with pytest.raises(
        ConnectFailedException, match="Could not connect to server: Connection failed"
    ):
        pigeon_client.connect(
            username=username, password=password, retry_limit=retry_limit
        )

    # Assert the logger was called the same number of times as the retry limit
    assert pigeon_client._logger.error.call_count == retry_limit


@pytest.mark.parametrize(
    "topic, data, expected_serialized_data",
    [
        ("topic1", {"field1": "value"}, '{"field1":"value"}'),
    ],
    ids=["send-data"],
)
def test_send(pigeon_client, topic, data, expected_serialized_data):
    expected_headers = {
        "source": pigeon_client._name,
        "service": "test",
        "hostname": pigeon_client._hostname,
        "pid": pigeon_client._pid,
        "hash": pigeon_client._hashes["topic1"],
        "sent_at": "1",
    }
    # Arrange
    with patch("pigeon.client.time.time_ns", lambda: 1e6):
        pigeon_client._topics[topic] = MockMessage
        pigeon_client._connection.send = MagicMock()

        # Act
        pigeon_client.send(topic, **data)

        # Assert
        pigeon_client._connection.send.assert_called_with(
            destination=topic, body=expected_serialized_data, headers=expected_headers
        )
        pigeon_client._logger.debug.assert_called_with(
            f"Sent data to {topic}: {expected_serialized_data}"
        )


@pytest.mark.parametrize(
    "topic, data",
    [
        ("unknown_topic", {"key": "value"}),
    ],
    ids=["send-data-no-such-topic"],
)
def test_send_no_such_topic(pigeon_client, topic, data):
    # Act & Assert
    with pytest.raises(NoSuchTopicException, match=f"Topic {topic} not defined."):
        pigeon_client.send(topic, **data)


@pytest.mark.parametrize(
    "topic, callback_name, expected_log",
    [
        ("topic1", "callback", "Subscribed to topic1 with {}."),
    ],
    ids=["subscribe-new-topic"],
)
def test_subscribe(pigeon_client, topic, callback_name, expected_log):
    # Arrange
    pigeon_client._topics[topic] = MockMessage
    callback = MagicMock(__name__=callback_name)
    pigeon_client._connection.subscribe = MagicMock()
    pigeon_client._update_state = MagicMock()

    # Act
    pigeon_client.subscribe(topic, callback)

    # Assert
    assert pigeon_client._callbacks[topic] == callback
    pigeon_client._connection.subscribe.assert_called_with(destination=topic, id=topic)
    pigeon_client._logger.info.assert_called_with(expected_log.format(callback))
    pigeon_client._update_state.assert_called()


@pytest.mark.parametrize(
    "topic",
    [
        ("unknown_topic"),
    ],
    ids=["subscribe-no-such-topic"],
)
def test_subscribe_no_such_topic(pigeon_client, topic):
    # Arrange
    callback = MagicMock()

    # Act & Assert
    with pytest.raises(NoSuchTopicException, match=f"Topic {topic} not defined."):
        pigeon_client.subscribe(topic, callback)


def test_subscribe_all(pigeon_client, mocker):
    pigeon_client._connection = mocker.MagicMock()
    pigeon_client.subscribe = mocker.MagicMock()

    callback = mocker.MagicMock()

    pigeon_client.subscribe_all(callback)

    assert len(pigeon_client.subscribe.mock_calls) == 2
    pigeon_client.subscribe.assert_any_call("topic1", callback, send_update=False)
    pigeon_client.subscribe.assert_any_call("topic2", callback, send_update=False)


def test_subscribe_all_with_core(pigeon_client, mocker):
    pigeon_client._connection = mocker.MagicMock()
    pigeon_client.subscribe = mocker.MagicMock()

    callback = mocker.MagicMock()

    pigeon_client.subscribe_all(callback, include_core=True)

    assert len(pigeon_client.subscribe.mock_calls) == 4
    pigeon_client.subscribe.assert_any_call("topic1", callback, send_update=False)
    pigeon_client.subscribe.assert_any_call("topic2", callback, send_update=False)
    pigeon_client.subscribe.assert_any_call(
        "&_announce_connection", callback, send_update=False
    )
    pigeon_client.subscribe.assert_any_call(
        "&_update_state", callback, send_update=False
    )


@pytest.mark.parametrize(
    "topic, expected_log",
    [
        ("topic1", "Unsubscribed from topic1."),
    ],
    ids=["unsubscribe-existing-topic"],
)
def test_unsubscribe(pigeon_client, topic, expected_log):
    # Arrange
    pigeon_client._callbacks[topic] = ["topic1"]
    pigeon_client._connection.unsubscribe = MagicMock()

    # Act
    pigeon_client.unsubscribe(topic)

    # Assert
    assert topic not in pigeon_client._callbacks
    pigeon_client._connection.unsubscribe.assert_called_with(id=topic)
    pigeon_client._logger.info.assert_called_with(expected_log)


def test_disconnect(pigeon_client):
    # Arrange
    pigeon_client._connection.is_connected = MagicMock(return_value=True)
    pigeon_client._connection.disconnect = MagicMock()
    pigeon_client._announce = MagicMock()

    # Act
    pigeon_client.disconnect()

    # Assert
    pigeon_client._connection.disconnect.assert_called_once()
    pigeon_client._logger.info.assert_called_with("Disconnected from STOMP server.")
    pigeon_client._announce.assert_called_with(connected=False)



from pigeon import BaseMessage
import json
from pydantic import ValidationError
import pytest


class Msg(BaseMessage):
    x: int
    y: int


def test_serialize():
    data = {"x": 1, "y": 2}
    msg = Msg(**data)

    assert json.loads(msg.serialize()) == data


def test_deserialize():
    data = {"x": 3, "y": 4}
    msg = Msg.deserialize(json.dumps(data))

    assert msg.x == data["x"]
    assert msg.y == data["y"]


def test_forbid_extra():
    with pytest.raises(ValidationError):
        Msg(x=1, y=2, z=3)
    with pytest.raises(ValidationError):
        Msg.deserialize(json.dumps({"x": 1, "y": 2, "z": 3}))


def test_mising_values():
    with pytest.raises(ValidationError):
        Msg(x=1)
    with pytest.raises(ValidationError):
        Msg.deserialize(json.dumps({"y": 3}))



from .client import Pigeon
import argparse
import yaml
from functools import partial
import json


class Listener:
    def __init__(self, disp_headers, record):
        self.message_received = False
        self.disp_headers = disp_headers
        self.record = record
        self.messages = []

    def callback(self, msg, topic, headers):
        if self.record:
            self.messages.append(
                {
                    "msg": msg.__dict__,
                    "topic": topic,
                    "headers": headers,
                }
            )
        else:
            print(f"Recieved message on topic '{topic}':")
            print(msg)
            if self.disp_headers:
                print("With headers:")
                for key, val in headers.items():
                    print(f"{key}={val}")
        self.message_received = True

    def write(self, path):
        with open(path, "w") as f:
            json.dump(self.messages, f)


def main():
    parser = argparse.ArgumentParser(prog="Pigeon CLI")
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="The message broker to connect to.",
    )
    parser.add_argument(
        "--port", type=int, default=61616, help="The port to use for the connection."
    )
    parser.add_argument(
        "--username",
        type=str,
        help="The username to use when connecting to the STOMP server.",
    )
    parser.add_argument(
        "--password",
        type=str,
        help="The password to use when connecting to the STOMP server.",
    )
    parser.add_argument(
        "-p", "--publish", type=str, help="The topic to publish a message to."
    )
    parser.add_argument(
        "-d", "--data", type=str, help="The YAML/JSON formatted data to publish."
    )
    parser.add_argument(
        "-s",
        "--subscribe",
        type=str,
        action="append",
        default=[],
        help="The topic to subscribe to.",
    )
    parser.add_argument(
        "-a", "--all", action="store_true", help="Subscribe to all registered topics."
    )
    parser.add_argument(
        "-r", "--record", type=str, help="Write received messages to a JSON file."
    )
    parser.add_argument(
        "--one", action="store_true", help="Exit after receiving one message."
    )
    parser.add_argument(
        "-l", "--list", action="store_true", help="List registered topics and exit."
    )
    parser.add_argument(
        "--headers", action="store_true", help="Display headers of received messages."
    )

    args = parser.parse_args()

    if args.list:
        connection = Pigeon("CLI")
        for topic in connection._topics:
            print(topic)
        return

    if args.publish is None and args.subscribe is None and not args.all:
        print("No action specified.")
        return

    if args.publish and args.data is None:
        print("Must also specify data to publish.")
        return

    if args.data and args.publish is None:
        print("Must also specify topic to publish data to.")
        return

    if args.record is not None and not (args.subscribe or args.all):
        print("No subscriptions to record.")
        return

    connection = Pigeon("CLI", args.host, args.port)
    connection.connect(args.username, args.password)

    if args.publish:
        connection.send(args.publish, **yaml.safe_load(args.data))

    if args.subscribe or args.all:
        listener = Listener(args.headers, args.record is not None)

    if args.all:
        connection.subscribe_all(listener.callback)
    else:
        for topic in args.subscribe:
            connection.subscribe(topic, listener.callback)

    if args.subscribe or args.all:
        try:
            while not (args.one and listener.message_received):
                pass
        except KeyboardInterrupt:
            if args.record is not None:
                listener.write(args.record)
            print("exiting")
    exit(0)


if __name__ == "__main__":
    main()
